/**
 * Codap communications module
 *
 * Additional layer of abstraction over the codap interface module to allow us
 * to keep our lengthy communications with codap out of the way.
 */

define([
    './lib/CodapInterface'],
  function (codapInterface, codapPluginConfig) {

    var CodapCom = function(getStateFunc, loadStateFunc) {
      this.codapConnected = false;
      this.itemProto = {};
      this.loadStateFunc = loadStateFunc;

      this.findOrCreateDataContext = this.findOrCreateDataContext.bind(this);
      this.error = this.error.bind(this);
      this.startNewExperimentInCODAP = this.startNewExperimentInCODAP.bind(this);
      this.addValuesToCODAP = this.addValuesToCODAP.bind(this);
      this.addMultipleSamplesToCODAP = this.addMultipleSamplesToCODAP.bind(this);
      this.openTable = this.openTable.bind(this);
      this.setDataSetName = this.setDataSetName.bind(this);
      this.findOrCreateDataContext = this.findOrCreateDataContext.bind(this);

      this.drawAttributes = null;
      this.collectionAttributes = null;

      codapInterface.on('get', 'interactiveState', getStateFunc);

      // this.init();
    };

    var appName = 'Sampler';
    var targetDataSetName = 'Sampler';
    function getTargetDataSetPhrase() { return  'dataContext[' + targetDataSetName + ']'; }

    CodapCom.prototype = {

      /*
       * Returns a Promise.
       */
      init: function() {
        // initialize the codapInterface
        return codapInterface.init({
          name: appName,
          title: appName,
          dimensions: {width: 235, height: 400},
          version: 'v0.4 (#' + window.codapPluginConfig.buildNumber + ')',
          preventDataContextReorg: false,
          stateHandler: this.loadStateFunc
        });
      },

      setDataSetName: function (name) {
        targetDataSetName = name;
      },

      findOrCreateDataContext: function () {
        this.codapConnected = true;
        // Determine if CODAP already has the Data Context we need.
        // If not, create it.
        return codapInterface.sendRequest({
            action:'get',
            resource: getTargetDataSetPhrase()
            }, function (result) {
            if (result && !result.success) {
              codapInterface.sendRequest({
                action: 'create',
                resource: 'dataContext',
                values: {
                  name: targetDataSetName,
                  collections: [
                    {
                      name: 'experiments',
                      attrs: [
                        {name: "experiment", type: 'categorical'},
                        {name: "sample_size", type: 'categorical'}
                      ],
                      childAttrName: "experiment"
                    },
                    {
                      name: 'samples',
                      parent: 'experiments',
                      // The parent collection has just one attribute
                      attrs: [{name: "sample", type: 'categorical'}],
                      childAttrName: "sample"
                    },
                    {
                      name: 'items',
                      parent: 'samples',
                      labels: {
                        pluralCase: "items",
                        setOfCasesWithArticle: "an item"
                      },
                      // The child collection also has just one attribute
                      attrs: [{name: "value"}]
                    }
                  ]
                }
              }, function(result) { console.log(result); });
            }
          }
        );
      },

      error: function(msg) {
        console.log(msg || "Failed to connect to CODAP");
        this.codapConnected = false;
      },

      startNewExperimentInCODAP: function(experimentNumber, sampleSize) {
        var _this = this;
        if (!_this.codapConnected) {
          console.log('Not in CODAP');
          return;
        }
        _this.openTable();

        this.itemProto = {
          experiment: experimentNumber,
          sample_size: sampleSize
        };
      },

      openTable: function() {
        if (!this.codapConnected) {
          return;
        }

        codapInterface.sendRequest({
          action: 'create',
          resource: 'component',
          values: {
            type: 'caseTable'
          }
        });
      },

      addMultipleSamplesToCODAP: function (samples, isCollector) {
        var _this = this;
        var items = [];
        samples.forEach(function (sample) {
          var run = sample.run;
          var item;
          sample.values.forEach(function(v) {
            if (!isCollector) {
              item = Object.assign({}, {sample: run, value: v}, _this.itemProto);
            } else {
              item = Object.assign({}, v, {sample: run}, _this.itemProto);
            }
            items.push(item);
          });
        });
        codapInterface.sendRequest({
          action: 'create',
          resource: getTargetDataSetPhrase() + '.item',
          values: items
        });
      },
      addValuesToCODAP: function(run, vals, isCollector) {
        this.addMultipleSamplesToCODAP([{run: run, values: vals}], isCollector);
      },

      deleteAll: function(device, populateContextsList) {
        var _this = this;
        codapInterface.sendRequest({
          action: 'delete',
          resource: getTargetDataSetPhrase() + '.collection[experiments].allCases'
        });
        var structure = { items: ['value']};
        Object.keys( structure).forEach( function( key) {
          var tValidAttrs = structure[ key];
          codapInterface.sendRequest( {
            action: 'get',
            resource: getTargetDataSetPhrase() + '.collection[' + key + '].attributeList'
          }).then( function( iResult) {
            var tMsgList = [];
            if( iResult.success) {
              iResult.values.forEach( function( iAttribute) {
                if( tValidAttrs.indexOf( iAttribute.name) < 0) {
                  tMsgList.push( {
                    action: 'delete',
                    resource: getTargetDataSetPhrase() + '.collection[' + key + '].attribute[' + iAttribute.name + ']'
                  });
                }
              });
              if( tMsgList.length > 0)
                codapInterface.sendRequest( tMsgList).then( function( iResult) {
                  if( iResult.success || (iResult.every( function(iItem) {
                        return iItem.success;
                      }))) {
                    if (device === "collector") {
                      return _this.getContexts().then(populateContextsList);
                    }
                  }
                  else {
                    return Promise.reject( "Failure to remove attributes");
                  }
                }).then( null, function( iMsg) {
                  console.log( iMsg);
                });
            }
          });
        });
      },

      getContexts: function() {
        var _this = this;
        return new Promise(function(resolve, reject) {
          if (!_this.codapConnected) {
            // we log that CODAP is not initiated. If we are in CODAP, it will
            // respond eventually.
            console.log('Not in CODAP');
          }

          codapInterface.sendRequest({
            action: 'get',
            resource: 'dataContextList'
          }, function(result) {
            if (result && result.success) {
              resolve(result.values);
            } else {
              resolve([]);
            }
          });
        });
      },

      setCasesFromContext: function(contextName, caseVariables) {
        var _this = this;
        return new Promise(function(resolve, reject) {

          function setCasesGivenCount (results) {
            if (results.success) {
              caseVariables = [];

              var count = results.values,
                  reqs = [];
              for (var i = 0; i < count; i++) {
                reqs.push({
                  action: 'get',
                  resource: _this.collectionResourceName + '.caseByIndex['+i+']'
                });
              }
              codapInterface.sendRequest(reqs).then(function(results) {
                results.forEach(function(res) {
                  caseVariables.push(res.values['case'].values);
                });
                resolve(caseVariables);
             });
            }
          }

          function addAttributes() {
            _this.collectionAttributes.forEach(function (attr) {
              if (_this.drawAttributes.indexOf(attr) < 0) {
                codapInterface.sendRequest({
                  action: 'create',
                  resource: getTargetDataSetPhrase() + '.collection[items].attribute',
                  values: [attr]
                });
              }
            });
          }

          function setCollection (result) {
            if (result.success) {
              _this.collectionResourceName = "dataContext[" + contextName + "].collection[" +
                  result.values[0].name + "]";
              codapInterface.sendRequest([
                {     // get the existing columns in the draw table
                  action: 'get',
                  resource: getTargetDataSetPhrase() + '.collection[items].attributeList'
                },
                {     // get the columns we'll be needing
                  action: 'get',
                  resource: _this.collectionResourceName
                },
                {     // get the number of cases
                  action: 'get',
                  resource: _this.collectionResourceName + '.caseCount'
                }
              ]).then(function(results) {
                _this.drawAttributes = results[0].values.map(function (res) {
                  return res.name;
                });
                _this.collectionAttributes = results[1].values.attrs.map(function (attr) {
                  // we will use the attribute definitions to create new attributes
                  // so we take the precaution of removing identifiers.
                  delete attr.id; delete attr.guid; return attr;
                });
                addAttributes();    // throw this over the wall
                return results[2];
              }).then(setCasesGivenCount);
            }
          }

          codapInterface.sendRequest({
            action: 'get',
            resource: 'dataContext[' + contextName + '].collectionList'
          }).then(setCollection);
        });
      },

      logAction: function( iMsg, iReplaceArgs) {
        codapInterface.sendRequest({
          action: 'notify',
          resource: 'logMessage',
          values: {
            formatStr: iMsg,
            replaceArgs: iReplaceArgs
          }
        });
      }
    };

    return CodapCom;
});
